#ifndef SCREEN_H_
#define SCREEN_H_

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

int getScreenWidth ();
int getScreenHeight ();

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SCREEN_H_ */
