/* stof.hpp
 * Copyright (C) 2020  Sven Jähnichen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef BUTILITIES_STOF_HPP_
#define BUTILITIES_STOF_HPP_

#include <string>
#include <stdexcept>

namespace BUtilities {

float stof (const std::string& str, size_t* idx = 0);

}

#endif /* BUTILITIES_STOF_HPP_ */
